LIBRARY IEEE;
USE IEEE.std_logic_1164.ALL;
USE IEEE.std_logic_arith.ALL;

ENTITY buffer50 IS
	PORT ( 	
			sad : OUT std_logic;		-- seal de paquete recibido que activa la segunda maquina de estados de la UC.
			reset : OUT std_logic;		-- seal de reset que afecta la primera maquina de estados de la unidad de control.
			clk : IN std_logic;			-- clk del programa funciona a 12 MHz
			rcv : IN std_logic;			-- seal de recepcion de datos que responde al la entrada 
										-- diferencial entre d+ y d-.
			vp : INOUT std_logic;		-- entrada de datos al sistema, si oe alto recibe sino envia.
			vm : INOUT std_logic;		-- igual que vp, trabajan juntos en modo diferencial.
			ebs : IN std_logic;			-- enable buffer salida.

			long_dadesb : IN integer RANGE 0 TO 255;		--  longitud del buffer de salida
			buffer_salida : IN std_logic_vector(0 to 255);	-- datos del buffer de salida					
			pidb : IN std_logic_vector(0 to 7);

			mode : OUT std_logic;		-- selecciona el modo de entrada de datos del sistema en 
										-- diferencial (alto) o singled-ended (bajo).
			oe : OUT std_logic;			-- a nivel alto enciende vp/vm como salidas y el rcv, a bajo son entradas
										-- en float desconecta la resistencia de pull-up
			renb : OUT std_logic;		-- a nivel alto rcv,vm y vp responden a las seales d+ y d- a bajo solo rcv 
										-- las otras estan en alta impedancia.
										-- con enumerate a bajo se desconecta la resistencia, a alto se conecta a 
										-- d+ o d- en funcion del modo de speed.
			susp : OUT std_logic;		-- en nivel alto modo suspendido, rcv es bajo d+ y d- en alta impedancia y 
										-- vp/vm son salidas activas si oe en alta impedancia.
			speed : OUT std_logic;		-- a nivel alto full-speed(12Mb/s) modo la resistencia se conecta a d+, sino 
										-- low-speed(1.5Mb/s) y la resistencia se conecta a d-.
			escriu : OUT std_logic;		-- seal de escritura en la memoria ram.
			buffer_pid : OUT std_logic_vector(0 to 7);		-- PID del paquete
			add_ram : OUT std_logic_vector(0 to 7);			-- direccion de la memoria RAM
			buffer_data : OUT std_logic_vector(0 to 63);	-- Datos para grabar en la memoria RAM
			crc_out : OUT std_logic;						-- resultado comprobacin de errores CRC
			bmRequestType : OUT std_logic_vector(0 to 7);	-- tipo peticion (request)
			bRequest : OUT std_logic_vector(0 to 7);		-- tipo peticion (request)
			wValue : OUT std_logic_vector(0 to 15);			-- aqui esta el valor de la direccio en el setup p.e.							
			wIndex : OUT std_logic_vector(0 to 15);			-- wIndex del paquete.
			wLength : OUT std_logic_vector(0 to 15);		-- wLength del paquete.
			length : OUT integer RANGE 0 TO 127;			-- longitud en entero de bytes a enviar/recibir, wLength numerico.
			llegir : OUT std_logic;							-- seal de lectura de la memoria RAM.
			led_5 : OUT std_logic;
			led_6 : OUT std_logic;
			led_7 : OUT std_logic;
			led_8 : OUT std_logic									
		);

-- Posibles PIDS que se pueden encoontrar.

	CONSTANT token_out	:	std_logic_vector(0 to 7) := ('1','0','0','0','0','1','1','1');
	CONSTANT token_in	:	std_logic_vector(0 to 7) := ('1','0','0','1','0','1','1','0');	
	CONSTANT sof		:	std_logic_vector(0 to 7) := ('1','0','1','0','0','1','0','1');
	CONSTANT setup		:	std_logic_vector(0 to 7) := ('1','0','1','1','0','1','0','0');
	CONSTANT data0		:	std_logic_vector(0 to 7) := ('1','1','0','0','0','0','1','1');
	CONSTANT data1		:	std_logic_vector(0 to 7) := ('1','1','0','1','0','0','1','0');
	CONSTANT data2		:	std_logic_vector(0 to 7) := ('1','1','1','0','0','0','0','1');
	CONSTANT mdata		:	std_logic_vector(0 to 7) := ('1','1','1','1','0','0','0','0');
	CONSTANT ackk		:	std_logic_vector(0 to 7) := ('0','1','0','0','1','0','1','1');
	CONSTANT nack		:	std_logic_vector(0 to 7) := ('0','1','0','1','1','0','1','0');
	CONSTANT stall		:	std_logic_vector(0 to 7) := ('0','1','1','1','1','0','0','0');
	CONSTANT nyet		:	std_logic_vector(0 to 7) := ('0','1','1','0','1','0','0','1');
	CONSTANT err		:	std_logic_vector(0 to 7) := ('0','0','1','1','1','1','0','0');
	CONSTANT split		:	std_logic_vector(0 to 7) := ('0','0','0','1','1','1','1','0');
	CONSTANT ping		:	std_logic_vector(0 to 7) := ('0','0','1','0','1','1','0','1');
	CONSTANT reserved	:	std_logic_vector(0 to 7) := ('0','0','0','0','1','1','1','1');
	

	CONSTANT ZERO		:	std_logic_vector(0 to 7) := ('0','0','0','0','0','0','0','0');	

	-- Valor inicial del residuo del CRC16
	CONSTANT UNOS		:	std_logic_vector(0 to 15) := ('1','1','1','1','1','1','1','1','1','1','1','1','1','1','1','1');	

	
	CONSTANT sync_pattern 	:	std_logic_vector(0 to 7) := ('0','1','0','1','0','1','0','0');	-- patron de sincronizacin.
	CONSTANT pol_gen 		:	std_logic_vector(0 to 4) := ('0','0','1','0','1');				-- polinomio generador CRC5
	CONSTANT comp_residuo	:	std_logic_vector(0 to 4) := ('0','1','1','0','0');				-- resultado final del CRC5 para comprobar
	CONSTANT pol_gen16		:	std_logic_vector(0 to 15) := ('1','0','0','0','0','0','0','0','0','0','0','0','0','1','0','1');
	CONSTANT comp_residuo16 :	std_logic_vector(0 to 15) := ('1','0','0','0','0','0','0','0','0','0','0','0','1','1','0','1');	

	CONSTANT width_data		:	INTEGER := 63;						-- maximo que queremos permitir de entrada de datos.
	CONSTANT width_sync_pattern		:	INTEGER := 7;
	CONSTANT width_pid		:	INTEGER := 7;
	CONSTANT width_address_ram	:	INTEGER := 7;	
	CONSTANT width_crc	:	INTEGER := 4;
	CONSTANT width_address	:	INTEGER := 10; 						-- 7  11 bits si incluye o no el endpoint.

	CONSTANT DEFAULT_DIR		: 	std_logic_vector(0 to 10) := ('0','0','0','0','0','0','0','0','0','0','0');

--	Estados del proceso que calcula el CRC
	
	TYPE stages IS (inici,comprovacion,es_correcto,fin);		

--	Estados del proceso de recepcin

	TYPE buf_states IS (inici,pid,dades,se0,fi,post_fi,graba);
	
--	Estados del proceso de envio

	TYPE buf_statesb IS (inici,pid_e,dades,crc_e,fi);

END buffer50;

ARCHITECTURE arq_buffer50 OF buffer50 IS
	
	SIGNAL estat_buf : buf_states;								-- controla el estado en la recepcin
	SIGNAL i : INTEGER RANGE 0 TO width_sync_pattern;			-- controla la longitud recibida de la seal SYNC 
	SIGNAL j : INTEGER RANGE 0 TO width_data;					-- controla el numero de datos recibidos
	SIGNAL k : INTEGER RANGE 0 TO width_pid;					-- controla la longitud del PID
	SIGNAL x : INTEGER;											-- controla paquetes recibidos en grupos de 64, es la direccin de la RAM para escritura
	SIGNAL y : INTEGER;											-- total de bits recibidos.
	SIGNAL bs : INTEGER RANGE 0 TO 6;							-- contador de unos para el bit sttufing

	SIGNAL vp_ant : std_logic;									-- Anterior valor de vp en la recepcin. Para la codificacin NRZI.
	SIGNAL envia : std_logic;									-- seal de envio/recepcion.
	SIGNAL buf_data : std_logic_vector(0 to width_data);		-- variable auxiliar para los datos recibidos .
	SIGNAL bp : std_logic_vector(0 to width_pid);				-- variable auxiliar para el pid.
	SIGNAL cont_reset : INTEGER;								-- contador de un reset.
	SIGNAL cont_no_reset : INTEGER;								-- contador para distinguir nuevo reset o el mismo.
																-- controla pequeas pausas en el reset.
	SIGNAL signal_incorrecta : INTEGER;							-- contador de seal incorrecta.
	SIGNAL cont_suspend : INTEGER;								-- controla la inactividad en el bus para el estado suspendido.
	SIGNAL cont_no_suspend : INTEGER;							-- controla la actividad de nuevo en el bus.
	SIGNAL suspend : std_logic;									-- indica el estado actual.
																-- contadores y seales al enviar
	SIGNAL ii : INTEGER;										-- Tiempo de espera antes de enviar. Un byte entre paquetes.
	SIGNAL ib : INTEGER RANGE 0 TO width_sync_pattern;			-- controla el bit de la seal SYNC enviada.
	SIGNAL kk : INTEGER RANGE 0 TO 15;							-- controla el pid del crc enviado.
	SIGNAL kb : INTEGER RANGE 0 TO 255;							-- controla el bit de los datos enviados.
	SIGNAL l : INTEGER RANGE 0 TO 5;							-- controla el bit del crc enviado.
	SIGNAL estat_bufb : buf_statesb;							-- controla el estado del envio.
	SIGNAL vpi : std_logic;										-- valor intermedio de la seal de salida vp que va a la placa
	SIGNAL vmi : std_logic;										-- valor intermedio de la seal de salida vm que va a la placa			
	
	SIGNAL vpant : std_logic;									-- Anterior valor de vp en el envio. Para la codificacin NRZI.
	SIGNAL vmant : std_logic;									-- Anterior valor de vm en el envio. Para la codificacin NRZI.
	
	SIGNAL ic : INTEGER RANGE 0 TO width_data;					-- Contador del bit que se esta comprobando en el crc5
	SIGNAL id : INTEGER RANGE 0 TO 255;							-- Contador del bit que se esta comprobando en el crc16
	SIGNAL crc16r : std_logic;									-- enciende comprobacin CRC16 en la recepcin de datos
	SIGNAL crc5 : std_logic;									-- enciende comprobacin CRC5 en la recepcin de datos
	SIGNAL crc16 : std_logic;									-- enciende la creacin del CRC16 en el envio de datos
	SIGNAL estat_act : stages;									-- controla el estado del proceso que calcula el crc
	SIGNAL remainder : std_logic_vector(0 to 15);				-- residuo del calculo
	SIGNAL bool_xor : std_logic;								-- XOR entre el siguiente bit y el bit de mayor peso del residuo.

	SIGNAL lengthw : integer RANGE 0 TO 127;					-- valor numerico del wLength
	SIGNAL e : integer RANGE 0 TO 127;							-- exponecial que ayuda en el calculo del wLength

	SIGNAL primer : std_logic;									-- variable para comprobaciones del programa
	SIGNAL crc : std_logic;										-- indica si el crc es correcto con un 1.
	SIGNAL yant : INTEGER;										-- indica el final del calculo del crc cuando alcanza el valor de bits recibidos.
	
BEGIN

	PROCESS(clk,ebs)
		BEGIN
		IF estat_bufb = fi AND l = 5 then 	-- en el clock 5 del estado fin de la mquina de envio acaba el proceso de enviar
			envia <= '0';						
		ELSIF rising_edge(ebs) then			-- con la seal ebs se activa el envio
			envia <= '1';
		END IF;
	END PROCESS;
	PROCESS(clk)
	BEGIN
		llegir <= '0';
	END PROCESS;	
	PROCESS(clk)
		BEGIN
		IF suspend = '1' THEN				-- control de las seales de salida a la placa
			vp <= 'Z';						-- su significado se explica en la memoria
			vm <= 'Z';
			oe <= 'Z';
			susp <= '1';
		ELSIF envia = '1' THEN				-- envio de datos
			vp <= vpi;
			vm <= vmi;
			vpant <= vpi;
			vmant <= vmi;
			oe <= '0';
			renb <= '0';
			susp <= '0';
			speed <= '1';
			mode <= '0';		
		ELSIF envia = '0' THEN				-- recepcion de datos
			vp <= 'Z';
			vm <= 'Z';
			oe <= '1';
			renb <= '1';
			susp <= '0';
			speed <= '1';
			mode <= '0';				
		END IF;
	END PROCESS;
	
	------------------------------------- PROCESOS DE LA CAPA FSICA -------------------------------------------
	------------------------------------------ RECIBIR -------------------------------------------------
	
	PROCESS(clk)
		BEGIN
		IF rising_edge(clk) THEN
				IF envia = '0' THEN
					CASE estat_buf IS
						WHEN inici =>						-- inicializamos variables
							estat_bufb <= inici;			-- se inicializa el ciclo de estados emisor en inicio
							l <= 0;							-- contador del se0 			
							crc16 <= '0';					-- apagado del crc16 enviar
							crc5 <= '0';					-- apagado del crc5 enviar	
							crc16r <= '0';					-- apagado del crc16 recibir							
							escriu <= '0';					-- seal de escritura para el ctrl de la ram
							e <= 1;							-- calculo del exponencial bit a bit
							j <= 0;
							x <= 0;
							y <= 0;
							k <= 0;		
							bs <= 0;									-- o quizas ya vale 2 aki											
							buffer_data <= "0000000000000000000000000000000000000000000000000000000000000000";																		
							buffer_pid <= reserved;	
------------------------------------------------------------------------------------------------------------
-- control de la seal de suspended
------------------------------------------------------------------------------------------------------------
							cont_no_suspend <= 0;
							IF suspend <= '0' THEN
								IF vp = '0' OR vm = '0' THEN	
									null;
								ELSE
									IF cont_suspend < 36000 THEN			
										cont_suspend <= cont_suspend + 1;
									ELSE
										led_7 <= '1';										
										suspend <= '1';
									END IF;
								END IF;
							ELSE
								IF (vp = '1' AND vm = '0') OR (vp = '0' AND vm = '1') THEN	
									IF cont_no_suspend < 240000 THEN				
										cont_no_suspend <= cont_no_suspend + 1;
									ELSE
										led_7 <= '0';		
										suspend <= '0';		
										cont_no_suspend <= 0;						
									END IF;
								END IF;
							END IF;							
------------------------------------------------------------------------------------------------------------
-- control de la seal de reset
------------------------------------------------------------------------------------------------------------

											-- cuando VP/VM esten durante 10 ms (minimo 2.5 us) a 0, 
											-- se considera un SE0, y se efectua un reset.
											-- El dispositivo pasa a un estado en el que esperara un 
											-- set_address para su configuracin.
   											-- suponiendo la frecuencia de clock de 24 Mhz equivale a 2'5 us
											-- a pesar que el reset del hub tiene que durar 50 ms en intervalos
											-- de no menos de 10 ms de seal correcta y no menos de 3 ms de
											-- incorrecta.
																-- control de la seal de reset
							IF cont_no_reset = 0 THEN			-- contador que indica si aun estamos en el mismo reset
								IF vp = '0' and vm = '0' THEN	-- SE0 seal que durante un periodo determinado indica un reset
									IF cont_reset > 30 then		-- contador del reset								
										reset <= '1';	
										sad <= '1';		
										cont_no_reset <= cont_no_reset + 1;	
										signal_incorrecta <= 0;		
									ELSE
										cont_reset <= cont_reset + 1;										
										reset <= '0';		
										sad <= '0';																		
									END IF;
								ELSE
									cont_reset <= 0;	
									reset <= '0';						
								END IF;		
							ELSE
								reset <= '0';	
								IF cont_reset < 600000 THEN				-- servira para calcular el tiempo total que puede llegar a 
																		-- durar un reset
									IF vp = '0' and vm = '0' THEN
										IF cont_no_reset < 120000 THEN	
											signal_incorrecta <= signal_incorrecta + 1;
											cont_no_reset <= cont_no_reset + 1;
											cont_reset <= cont_reset + 1;												
										ELSE
											signal_incorrecta <= 0;
											cont_reset <= cont_reset + 1;												
										END IF;
									ELSE
										IF signal_incorrecta < 36000 THEN
											signal_incorrecta <= signal_incorrecta + 1;
											cont_no_reset <= 1;
											cont_reset <= cont_reset + 1;												
										ELSE 
											cont_no_reset <= 0;
											cont_reset <= 0;	
										END IF;
									END IF;
								ELSE
									cont_reset <= 0;	
									cont_no_reset <= 0;
								END IF;
							END IF;		
------------------------------------------------------------------------------------------------------------
-- control de la seal de sincronizacin SYNC
------------------------------------------------------------------------------------------------------------							
							IF vp = sync_pattern(i) AND suspend = '0' THEN	
								IF i = width_sync_pattern THEN	-- si se coincide con el patron durante 8 ciclos se ha completado su recepcin
									estat_buf <= pid;			-- pasamos al siguiente estado
									vp_ant <= '0';		
									lengthw <= 0;	
									i <= 0;
									ii <= 0;					-- contador de espera antes de enviar lo inicializamos aqu
								ELSE
									i <= i + 1;
								END IF;
							ELSIF vp = sync_pattern(0) THEN
								i <= 1;
							ELSE 
								i <= 0;
							END IF;
						WHEN pid =>								-- Se recoge el PID.
							sad <= '0';		
							IF vp = '0' AND vm = '0' THEN
								estat_buf <= inici;							
--								led_5 <= '0';					-- error. se0 no se puede dar aun.
							ELSIF vp = '1' AND vm = '1' THEN
								estat_buf <= inici;							
--								led_5 <= '0';					-- error. no se puede dar.
							ELSE
								IF bs < 6 THEN
									IF vp_ant = vp THEN			-- NRZI encoding cambio de nivel implica 0 lgico.
										buffer_pid(k) <= '1';	-- se almacena en las variables
										bp(k) <= '1';
										bs <= bs + 1;
									ELSIF vp_ant /= vp THEN											
										buffer_pid(k) <= '0';
										bp(k) <= '0';											
										bs <= 0;
									END IF;
									IF k < width_pid THEN
										k <= k + 1;
									ELSE
										estat_buf <= dades;		-- A los 8 bits salta a datos
										k <= 0;
									END IF;
								ELSE							-- bit stuffing, no recogemos el 0.
									bs <= 0;
									IF vp_ant = vp THEN
										estat_buf <= inici;										
--										led_5 <= '0';			-- error. se supone que aqui esta obligado un 0.
									END IF;										
								END IF;		
								vp_ant <= vp;
							END IF;
						WHEN dades =>							-- Se recogen los datos
							IF vp = '0' AND vm = '0' THEN
								estat_buf <= se0;				-- Al recibir la seal de SE0 se pasa al siguiente estado
								escriu <= '0';		
							ELSIF vp = '1' AND vm = '1' THEN
								estat_buf <= inici;
--								led_5 <= '0';					-- error
							ELSE
								IF bp = setup OR bp = token_in OR bp = token_out THEN	-- se activa el crc5
										crc5 <= '1';	
								ELSIF bp = data0 OR bp = data2 OR bp = mdata OR bp = data1 then	-- se activa el crc16
										crc16r <= '1';
								END IF;
								IF bs < 6 THEN					-- controla el bit sttuffing.
									IF vp_ant = vp THEN			-- NRZI encoding cambio de nivel implica 0 lgico.
										bs <= bs + 1;
										buf_data(j) <= '1';
										buffer_data(j) <= '1';		
										IF j >= 48 THEN			-- Se calcula el wLength situado a partir de la posicin 48
											e <= e + e;			-- valor a sumar si el bit es 1.
											lengthw <= lengthw + e;	
										END IF;
									ELSIF vp_ant /= vp THEN										
										bs <= 0;
										buf_data(j) <= '0';
										buffer_data(j) <= '0';
										IF j >= 48 THEN			-- si es cero no se suma.
											e <= e + e;
										END IF;										
									END IF;

									y <= y + 1;					-- contador de bits recibidos

									IF j = width_data THEN		-- controla el bit dentro del paquete de 64
										j <= 0;					
										x <= x + 1;				-- controla el numero de paquetes							
										escriu <= '1';		
									ELSIF j = width_data - 1 THEN		
										add_ram <= conv_std_logic_vector(x,8);	-- se transforma x en la direccin de la RAM
										j <= j + 1;
										escriu <= '0';							-- seal de grabacin en la RAM
									ELSIF (j = 0 AND x /= 0) THEN				-- justo al final del primer paquete se guardan la variables
																				-- si se trata del data0 que acompaa al SETUP ser muy til 
																				-- tenerlas almacenadas
										j <= j + 1;								
										escriu <= '0';							
										bmRequestType(0 to 7) <= buf_data(0 to 7);
										bRequest(0 to 7) <= buf_data(8 to 15);
										wValue(0 to 15) <= buf_data(16 to 31);
										wIndex(0 to 15) <= buf_data(32 to 47);
										wLength(0 to 15) <= buf_data(48 to 63);		
										led_8 <= '1';																			
--										IF	address_ram = ADR THEN			-- error. sobrepasada capacidad				
--											led_5 <= '0';
--										END IF;
									ELSE									
										j <= j + 1;
										escriu <= '0';											
									END IF;
								ELSE											-- bit stuffing, no recogemos el 0.
									IF vp_ant /= vp THEN
										bs <= 0;
									ELSIF vp_ant = vp THEN
										estat_buf <= inici;										
--										led_5 <= '0';				-- error. No se cumple el bit stuffing.
									END IF;
									escriu <= '0';										
								END IF;
								vp_ant <= vp;									
							END IF;	
						WHEN se0 =>								
							IF vp = '0' AND vm = '0' THEN				-- segundo bit del SE0
																		-- son 2 clocks a 12 MHz, uno en el estado anterior.
								escriu <= '0';						
								estat_buf <= fi;						-- se pasa al estado fi automticamente.
								length <= lengthw;					
								IF bp = setup OR bp = token_in OR bp = token_out then	
									wValue(0 to 10) <= buf_data(0 to 10); 	-- se almacena la direccin aprovechando el buffer
																			-- wValue en los paquetes de testigo.
								END IF;
							ELSE	
								estat_buf <= inici;
--								led_5 <= '0';							-- error.
							END IF;
						WHEN fi =>					
							cont_no_reset <= 0;							-- se resetean estas variables de los valores residuales 
							cont_reset <= 0;							-- que tengan.		
							cont_suspend <= 0;
							IF bp = setup AND estat_act = fin THEN		-- Cuando el paquete es setup y se termina la comprobacin.
								crc5 <= '0';							-- se apaga la maquina del crc
							    IF crc = '1' AND y = 16 THEN 			-- se comprueba antes de enviar la seal de paquete recibido.
   									estat_buf <= post_fi;	
									sad <= '1';		
									led_5 <= '1';										   									
								ELSE
									estat_buf <= inici;										
								END IF;
							ELSIF bp = token_in AND estat_act = fin THEN	
								crc5 <= '0';	
							    IF crc = '1' AND y = 16 THEN 								
									estat_buf <= post_fi;	
									sad <= '1';	
								ELSE
									estat_buf <= inici;	
								END IF;
							ELSIF bp = token_out AND estat_act = fin THEN	
								crc5 <= '0';		
								IF crc = '1' AND y = 16 THEN 								
									estat_buf <= post_fi;	
									sad <= '1';	
								ELSE
									estat_buf <= inici;	
								END IF;
							ELSIF bp = data0 AND estat_act = fin THEN		
								crc16r <= '0';			
								IF crc = '1' THEN
									estat_buf <= graba;				-- en los paquetes de datos antes de finalizar 
									led_6 <= '1';					-- se graban los ltimos bits en el estado graba.
								ELSE
									estat_buf <= inici;	
								END IF;																
								IF j > 4 THEN						-- si el ltimo paquete es muy pequeo se inicializa
									j <= 2;							-- esta variable a 0 para que el proceso graba espere 
								ELSE								-- un poco ms antes de empezar.
									j <= 0;
								END IF; 								
							ELSIF bp = data1 and estat_act = fin THEN	
								crc16r <= '0';
								IF crc = '1' THEN
									estat_buf <= graba;	
								ELSE
									estat_buf <= inici;								
								END IF;									
								IF j > 4 THEN				
									j <= 2;					
								ELSE
									j <= 0;
								END IF; 								
							ELSIF bp = setup  OR bp = token_in OR bp = token_out THEN	-- se espera a que termine el crc
								null;							
							ELSIF bp = data0 OR bp = data1 OR bp = data2 OR bp = mdata THEN	
								null;
							ELSIF bp = ackk THEN
								estat_buf <= post_fi;	
								sad <= '1';																	
							ELSE
								estat_buf <= post_fi;	
								sad <= '1';								
							END IF;									
						WHEN graba =>						
							IF j < 3 THEN								-- Hay que retrasar esta grabacion por si se 
								j <= j + 1;								-- realiza muy seguida a la ltima.
							ELSIF j < 4 THEN								
								j <= j + 1;							
								add_ram <= conv_std_logic_vector(x,8);									
							ELSE
								j <= 0;
								escriu <= '1';
								estat_buf <= post_fi;
								sad <= '1';								
							END IF;		
						WHEN post_fi =>									-- para desactivar este dispositivo						
																		-- desconecta eb.
							escriu <= '0';		
							sad <= '0';									-- la seal de envio sad vuelve a 0 antes de empezar de nuevo.
							estat_buf <= inici;								
					END CASE;				
				ELSE
				
	------------------------------------------ ENVIAR --------------------------------------------------				
					CASE estat_bufb IS
						WHEN inici =>
							estat_buf <= inici;		
							if ii < 6 then	-- hay que dejar minimo 8 bits antes de enviar, con 6 en este momento son suficientes.
									ii <= ii + 1;	
									crc16 <= '0';	-- se apagan las maquinas de crc encendidas
									crc5 <= '0';			
									crc16r <= '0';										
							else
								IF pidb = data0 OR pidb = data1 OR pidb = data2 OR pidb = mdata THEN
									crc16 <= '1';
								END IF;
								i <= 0;	
								l <= 0;									-- seal de fin de SE0
								k <= 0;									-- contador del bit de pid
								kb <= 0;								-- contador del bit de datos
								kk <= 0;								-- contador del bit del crc
								vpi <= sync_pattern(ib);
								vmi <= '0';
								IF ib < width_sync_pattern THEN			-- se envia el paquete de sincronizacin
									ib <= ib + 1;						-- contador del bit del SYNC
								ELSE
									ii <= 0;
									estat_bufb <= pid_e;		
									bs <= 1;							-- contador bit sttuffing 			
									ib <= 0;							-- cuando no hay cambio de nivel p.e. 2 zeros seguidos
								END IF;									-- se da al final del SYNC por lo que bs = 1
							end if;
						WHEN pid_e =>									-- Se envia el PID
							IF bs < 6 THEN
								IF pidb(k) = '0' THEN
									vpi <= not vpi;						-- Segun el NRZI el bit 0 es un cambio en la linea.
									vmi<= '0';							-- En este modo de envio vmi siempre es 0 excepto en el SE0
									bs <= 0;								
								ELSE 
									vpi <= vpi;
									vmi<= '0';
									bs <= bs + 1;	
								END IF;
								IF k < width_pid THEN				
									k <= k + 1;		
								ELSE
									k <= 0;
									IF long_dadesb > 0 THEN				-- si hay datos se pasa al estado para su envio
										estat_bufb <= dades;			
									ELSE								-- si no hay datos se salta al crc en un paquete de datos
										IF pidb = data0 OR pidb = data1 OR pidb = data2 OR pidb = mdata THEN
											estat_bufb <= crc_e;		
											crc16 <= '0';				-- se apaga la mquina
										ELSE
											estat_bufb <= fi;			-- se salta al fi en paquetes de handshake.
										END IF;
									END IF;
								END IF;
							ELSE										-- bit sttuffing	
								bs <= 0;
								IF vpi = '0' THEN
									vpi <= '1';
									vmi <= '0';
								ELSE
									vpi <= '0';
									vmi <= '0';
								END IF;								
							END IF;
						WHEN dades =>
							IF bs < 6 THEN
								IF buffer_salida(kb) = '0' THEN 		-- solo cambia si es 0
									IF vpi = '0' THEN
										vpi <= '1';
										vmi <= '0';
									ELSE
										vpi <= '0';
										vmi<= '0';
									END IF;
									bs <= 0;
								ELSE
									vpi <= vpi;
									vmi <= '0';
									bs <= bs + 1;
								END IF;
								IF kb < long_dadesb THEN
									kb <= kb + 1;
								ELSE 
									kb <= 0;
									estat_bufb <= crc_e;			
									crc16 <= '0';
								END IF;
							ELSE										-- bit sttuffing
								bs <= 0;
								IF vpi = '0' THEN
									vpi <= '1';
									vmi <= '0';
								ELSE
									vpi <= '0';
									vmi <= '0';
								END IF;								
							END IF;
						WHEN crc_e =>
							IF bs < 6 THEN
								IF remainder(kk) = '0' THEN 			-- solo cambia si es 0
									IF vpi = '1' THEN					-- Aqu hay que enviar el remainder
										vpi <= '1';						-- invertido por lo que un zero es un uno 
										vmi <= '0';						-- realmente, cambio vpi = '0' por vpi = '1'
									ELSE								-- en la condicion del if
										vpi <= '0';
										vmi <= '0';
									END IF;
									bs <= bs + 1;						-- sumo con el 0 porque en realidad es un uno		
								ELSE									-- se invierte
									vpi <= not vpi;
									vmi <= '0';							-- y pongo el not en el vpi
									bs <= 0;
								END IF;					
								IF kk < 15 THEN
									kk <= kk + 1;
								ELSE 
									estat_bufb <= fi;	
									kk <= 0;
								END IF;
							ELSE
								bs <= 0;
								IF vpi = '0' THEN
									vpi <= '1';
									vmi <= '0';
								ELSE
									vpi <= '0';
									vmi <= '0';
								END IF;								
							END IF;
						WHEN fi =>									-- estat fi crea el SE0 durante 2*periodo.
							IF bs < 6 THEN						    -- el bit stuffing
								IF l < 1 THEN	
									primer <= '1';
									l <= l + 1;
									vpi <= '0';
									vmi <= '1';
								ELSIF l < 3 THEN
									l <= l + 1;
									vpi <= '0';
									vmi <= '1';
								ELSIF l < 5 THEN			-- experimentalmente se comprueva que el SE0 debe durar 4 ciclos
--									led_8 <= '1';			-- con 2 que seria la teoria no son suficientes.					
									l <= l + 1;
									vpi <= '1';
									vmi <= '0';									
								ELSE
								    l <= 0;
								END IF;
							ELSE
								bs <= 0;
								IF vpi = '0' THEN
									vpi <= '1';
									vmi <= '0';
								ELSE
									vpi <= '0';
									vmi <= '0';
								END IF;								
							END IF;
					END CASE;
				END IF;
		END IF;
	END PROCESS;

	-------------------------------------- CONTROLES DE ERROR -------------------------------------------	
	-------------------------------------------- CRC 5 --------------------------------------------------	

	PROCESS(clk)			
		BEGIN
		IF rising_edge(clk) THEN
			IF crc5 = '1' THEN 												
				CASE estat_act IS											
					WHEN inici =>
						ic <= 1;						-- indice del bit comprobado
						yant <= 1;						-- contador del numero de bits que llevamos para finalizar el proceso 
						crc <= '0';						-- resultado del crc para comprobaciones internas del programa.
						crc_out <= '0';					-- seal de salida del modulo con el resultado.
						remainder(0) <= '1';			-- se inicializa el residuo
						remainder(1) <= '1';
						remainder(2) <= '1';
						remainder(3) <= '1';
						remainder(4) <= '1';
						estat_act <= comprovacion;		-- se pasa al siguiente estado conociendo ya el primer bool_xor
						bool_xor <= '1' XOR buf_data(0);
					WHEN comprovacion =>
						IF bs < 6 THEN
							IF bool_xor = '0' THEN
--								para reducir el tiempo seguimos optimizando
--								en vez de volver a inici ya calculamos aki.
								remainder(0) <= remainder(1);
								remainder(1) <= remainder(2);
								remainder(2) <= remainder(3);
								remainder(3) <= remainder(4);
								remainder(4) <= '0';
								IF y = yant THEN		-- cuando el contador alcance el valor de bits recibido se salta a la comprobacin.
									estat_act <= es_correcto;	
								ELSE
									bool_xor <= remainder(1) XOR buf_data(ic);	-- en cada estado se avanza con el calculo del siguiente bool									
									ic <= ic + 1;
									yant <= yant + 1;								
								END IF;								
							ELSIF bool_xor = '1' THEN	-- si el bool_xor es 1 se debe hacer la xor con el polinomio generador.
								remainder(0) <= remainder(1) XOR pol_gen(0);
								remainder(1) <= remainder(2) XOR pol_gen(1);
								remainder(2) <= remainder(3) XOR pol_gen(2);
								remainder(3) <= remainder(4) XOR pol_gen(3);
								remainder(4) <= '0' XOR pol_gen(4);
								IF y = yant THEN
									estat_act <= es_correcto;	
								ELSE							
									bool_xor <= ((remainder(1) XOR pol_gen(0)) XOR buf_data(ic));	
									ic <= ic + 1;
									yant <= yant + 1;								
								END IF;				
							END IF;
						END IF;
					WHEN es_correcto =>
						IF comp_residuo(0 to 4) = remainder(0 to 4) THEN	-- si el residuo coincide con su comprobacin el crc es correcto
							crc <= '1';	
							crc_out <= '1';
						END IF;														
						estat_act <= fin;
					WHEN fin =>
						null;
				END CASE;

	-------------------------------------------- CRC 16 ENVIAR -----------------------------------------------
				
			ELSIF crc16 = '1' THEN		-- idem que el crc5 pero con 16 bits, al final no se comprueba porque este ser el crc enviado.
				CASE estat_act IS
					WHEN inici =>
						id <= 1;
						remainder(0) <= '1';
						remainder(1) <= '1';
						remainder(2) <= '1';
						remainder(3) <= '1';
						remainder(4) <= '1';
						remainder(5) <= '1';
						remainder(6) <= '1';
						remainder(7) <= '1';
						remainder(8) <= '1';
						remainder(9) <= '1';
						remainder(10) <= '1';
						remainder(11) <= '1';
						remainder(12) <= '1';
						remainder(13) <= '1';
						remainder(14) <= '1';
						remainder(15) <= '1';
						IF long_dadesb = 0 THEN
							estat_act <= es_correcto;
						ELSE
							estat_act <= comprovacion ;		
							bool_xor <= '1' XOR buffer_salida(0);	
						END IF;
					WHEN comprovacion =>
						IF bool_xor = '0' THEN
							remainder(0) <= remainder(1);
							remainder(1) <= remainder(2);
							remainder(2) <= remainder(3);
							remainder(3) <= remainder(4);
							remainder(4) <= remainder(5);
							remainder(5) <= remainder(6);
							remainder(6) <= remainder(7);
							remainder(7) <= remainder(8);
							remainder(8) <= remainder(9);
							remainder(9) <= remainder(10);
							remainder(10) <= remainder(11);
							remainder(11) <= remainder(12);
							remainder(12) <= remainder(13);
							remainder(13) <= remainder(14);
							remainder(14) <= remainder(15);
							remainder(15) <= '0';
							IF id = 0 THEN
								estat_act <= es_correcto;								
							ELSIF id < long_dadesb THEN
								bool_xor <= remainder(1) XOR buffer_salida(id);							
								id <= id + 1;
							ELSE
								bool_xor <= remainder(1) XOR buffer_salida(id);			
								id <= 0;							
							END IF;								
						ELSIF bool_xor = '1' THEN
							remainder(0) <= remainder(1) XOR pol_gen16(0);
							remainder(1) <= remainder(2) XOR pol_gen16(1);
							remainder(2) <= remainder(3) XOR pol_gen16(2);
							remainder(3) <= remainder(4) XOR pol_gen16(3);
							remainder(4) <= remainder(5) XOR pol_gen16(4);
							remainder(5) <= remainder(6) XOR pol_gen16(5);
							remainder(6) <= remainder(7) XOR pol_gen16(6);
							remainder(7) <= remainder(8) XOR pol_gen16(7);
							remainder(8) <= remainder(9) XOR pol_gen16(8);
							remainder(9) <= remainder(10) XOR pol_gen16(9);
							remainder(10) <= remainder(11) XOR pol_gen16(10);
							remainder(11) <= remainder(12) XOR pol_gen16(11);
							remainder(12) <= remainder(13) XOR pol_gen16(12);
							remainder(13) <= remainder(14) XOR pol_gen16(13);
							remainder(14) <= remainder(15) XOR pol_gen16(14);
							remainder(15) <= '0' XOR pol_gen16(15);
							IF id = 0 THEN
								estat_act <= es_correcto;	
							ELSIF id < long_dadesb THEN
								bool_xor <= ((remainder(1) XOR pol_gen16(0)) XOR buffer_salida(id));	
								id <= id + 1;
							ELSE
								bool_xor <= ((remainder(1) XOR pol_gen16(0)) XOR buffer_salida(id));	
								id <= 0;		
							END IF;				
						END IF;
					WHEN es_correcto =>
						estat_act <= fin;
					WHEN fin =>
						null;
				END CASE;
				
	-------------------------------------------- CRC 16 RECIBIR ---------------------------------------------	
				
		ELSIF crc16r = '1' THEN
				CASE estat_act IS
					WHEN inici =>		
						ic <= 1;
						yant <= 1;
						crc <= '0';				
						crc_out <= '0';	
						remainder(0) <= '1';
						remainder(1) <= '1';
						remainder(2) <= '1';
						remainder(3) <= '1';
						remainder(4) <= '1';
						remainder(5) <= '1';
						remainder(6) <= '1';
						remainder(7) <= '1';
						remainder(8) <= '1';
						remainder(9) <= '1';
						remainder(10) <= '1';
						remainder(11) <= '1';
						remainder(12) <= '1';
						remainder(13) <= '1';
						remainder(14) <= '1';
						remainder(15) <= '1';
						estat_act <= comprovacion;	
						bool_xor <= '1' XOR buf_data(0);	
					WHEN comprovacion =>
						IF bs < 6 THEN									-- si bs = 6 se para un turno.
							IF bool_xor = '0' THEN
								remainder(0) <= remainder(1);
								remainder(1) <= remainder(2);
								remainder(2) <= remainder(3);
								remainder(3) <= remainder(4);
								remainder(4) <= remainder(5);
								remainder(5) <= remainder(6);
								remainder(6) <= remainder(7);
								remainder(7) <= remainder(8);
								remainder(8) <= remainder(9);
								remainder(9) <= remainder(10);
								remainder(10) <= remainder(11);
								remainder(11) <= remainder(12);
								remainder(12) <= remainder(13);
								remainder(13) <= remainder(14);
								remainder(14) <= remainder(15);
								remainder(15) <= '0';
								IF y = yant THEN
									estat_act <= es_correcto;	
								ELSE
									bool_xor <= remainder(1) XOR buf_data(ic);							
									ic <= ic + 1;
									yant <= yant + 1;									
								END IF;								
							ELSIF bool_xor = '1' THEN
								remainder(0) <= remainder(1) XOR pol_gen16(0);
								remainder(1) <= remainder(2) XOR pol_gen16(1);
								remainder(2) <= remainder(3) XOR pol_gen16(2);
								remainder(3) <= remainder(4) XOR pol_gen16(3);
								remainder(4) <= remainder(5) XOR pol_gen16(4);
								remainder(5) <= remainder(6) XOR pol_gen16(5);
								remainder(6) <= remainder(7) XOR pol_gen16(6);
								remainder(7) <= remainder(8) XOR pol_gen16(7);
								remainder(8) <= remainder(9) XOR pol_gen16(8);
								remainder(9) <= remainder(10) XOR pol_gen16(9);
								remainder(10) <= remainder(11) XOR pol_gen16(10);
								remainder(11) <= remainder(12) XOR pol_gen16(11);
								remainder(12) <= remainder(13) XOR pol_gen16(12);
								remainder(13) <= remainder(14) XOR pol_gen16(13);
								remainder(14) <= remainder(15) XOR pol_gen16(14);
								remainder(15) <= '0' XOR pol_gen16(15);
								IF y = yant THEN
									estat_act <= es_correcto;
								ELSE
									bool_xor <= ((remainder(1) XOR pol_gen16(0)) XOR buf_data(ic));	
									ic <= ic + 1;
									yant <= yant + 1;									
								END IF;				
							END IF;
						END IF;
					WHEN es_correcto =>				
						if comp_residuo16(0 to 15) = remainder(0 to 15) then
							crc <= '1';	
							crc_out <= '1';
						end if;		
						estat_act <= fin;										
					WHEN fin =>
						null;						
				END CASE;
			ELSE 
				estat_act <= inici;
			END IF;
		END IF;
	END PROCESS;
END arq_buffer50;
